local dialog = nil
local vb = nil

class "RenoiseScriptingTool" (renoise.Document.DocumentNode)
function RenoiseScriptingTool:__init()    
  renoise.Document.DocumentNode.__init(self) 
  self:add_property("Name", "Untitled Tool")
  self:add_property("Version", "1.0")
end

local manifest = RenoiseScriptingTool()
local ok,err = manifest:load_from("manifest.xml")
local tool_name = manifest:property("Name").value
local tool_version = manifest:property("Version").value

prefs = renoise.Document.create {
  pattern_length = 64
}
renoise.tool().preferences = prefs

-- Constants.
local MAX_PATTERN_LINES = renoise.Pattern.MAX_NUMBER_OF_LINES
local NOTE_OFF = (renoise.API_VERSION == 4) and renoise.PatternLine.NOTE_OFF or renoise.PatternTrackLine.NOTE_OFF
local TRACK_TYPE_SEQUENCER = renoise.Track.TRACK_TYPE_SEQUENCER

-- Functions.
local function round(value)
  return math.floor(value + 0.5)
end

function safe_pattern_length(value)
  return math.min(MAX_PATTERN_LINES, math.max(1, value))
end

function adjust_pattern_length(factor)
  return safe_pattern_length(round(renoise.song().selected_pattern.number_of_lines * factor))
end

function resize_pattern(pattern, new_length)
  
  -- We need a valid pattern object.
  if (pattern == nil) then
    renoise.app():show_status('Need a valid pattern object!')
    return
  end
  
  -- Shortcut to the song object.
  local rs = renoise.song()
  
  -- Get the current pattern length.
  local src_length = pattern.number_of_lines 
  
  -- Make sure new_length is within valid limits.
  local dst_length = math.min(MAX_PATTERN_LINES, math.max(1, new_length))
   
  -- If the new length is the same as the old length, then we have nothing to do.
  if (dst_length == src_length) then
    return
  end
  
  -- Set conversion ratio.
  local ratio = dst_length / src_length
  
  -- Change pattern length.
  pattern.number_of_lines = dst_length
   
  -- Source.
  local src_track = nil
  local src_line = nil
  local src_note_column = nil
  local src_effect_column = nil
  
  -- Insert a new track as a temporary work area.
  rs:insert_track_at(1)
  
  -- Destination.
  local dst_track = pattern:track(1)
  local dst_line_index = 0
  local dst_delay = 0
  local dst_line = nil
  local dst_note_column = nil
  local dst_effect_column = nil
  
  -- Misc.
  local tmp_line_index = 0
  local tmp_line_delay = 0
  local delay_column_used = false   
  local track = nil

  -- Iterate through each track.
  for src_track_index = 2, #rs.tracks, 1 do
  
    track = rs:track(src_track_index)

    -- Set source track.
    src_track = pattern:track(src_track_index)
    
    -- Reset delay check.
    delay_column_used = false
 
    -- Iterate through source lines.
    for src_line_index = 0, src_length - 1, 1 do
    
      -- Set source line.
      src_line = src_track:line(src_line_index + 1)
      
      -- Only process source line if it contains data.
      if (not src_line.is_empty) then
           
        -- Store temporary line index and delay.
        tmp_line_index = math.floor(src_line_index * ratio)
        tmp_line_delay = math.floor(((src_line_index * ratio) - tmp_line_index) * 256)
         
        -- Process note columns.
        for note_column_index = 1, track.visible_note_columns, 1 do
        
          -- Set source note column.
          src_note_column = src_line:note_column(note_column_index)
          
          -- Only process note column if it contains data.
          if (not src_note_column.is_empty) then
          
            -- Calculate destination line and delay.
            dst_line_index = tmp_line_index
            dst_delay = math.ceil(tmp_line_delay + (src_note_column.delay_value * ratio))
            
            -- Wrap note to next line if necessary.
            while (dst_delay >= 256) do
              dst_delay = dst_delay - 256
              dst_line_index = dst_line_index + 1
            end
            
            -- Keep track of whether the delay column is used
            -- so that we can make it visible later if necessary.
            if (dst_delay > 0) then
              delay_column_used = true
            end
            dst_line = dst_track:line(dst_line_index + 1)
            dst_note_column = dst_line:note_column(note_column_index)
            
            -- Note prioritisation.
            if (dst_note_column.is_empty) then
            
              -- Destination is empty. Safe to copy.
              dst_note_column:copy_from(src_note_column)
              dst_note_column.delay_value = dst_delay   
              
            else
              -- Destination contains data. 
              -- Try to prioritise which note to keep...
            
              -- If destination contains a note-off...
              if (dst_note_column.note_value == NOTE_OFF) then
              
                -- Source note takes priority.
                dst_note_column:copy_from(src_note_column)
                dst_note_column.delay_value = dst_delay
                
              else
              
                -- If the source note is less delayed, or louder, than the
                -- destination note...
                if (src_note_column.delay_value < dst_note_column.delay_value) 
                or (src_note_column.volume_value > dst_note_column.volume_value) then
                
                  -- Source note takes priority.
                  dst_note_column:copy_from(src_note_column)
                  dst_note_column.delay_value = dst_delay
                  
                end
                
              end      
              
            end -- End: Note prioritisation.
          
          end -- End: Only process note column if it contains data.
         
        end -- End: Process note columns.
          
        -- Process effect columns.
        for effect_column_index = 1, track.visible_effect_columns, 1 do
          src_effect_column = src_line:effect_column(effect_column_index)
          if (not src_effect_column.is_empty) then
            dst_effect_column = dst_track:line(round(src_line_index * ratio) + 1):effect_column(effect_column_index)
            if (dst_effect_column.is_empty) then
              dst_effect_column:copy_from(src_effect_column)
            end
          end
        end
      
      end -- End: Only process source line if it contains data.

    end -- End: Iterate through source lines.
    
    -- If there is automation to process...
    if (#src_track.automation > 0) then
    
      -- Copy processed lines from temporary track back to original track.
      -- We can't simply use dst_track:copy_from(src_track) here, since 
      -- it will erase the automation.
      for line_index = 1, dst_length, 1 do
        dst_line = dst_track:line(line_index)
        src_line = src_track:line(line_index)
        src_line:copy_from(dst_line)
      end
    
      -- Process automations.
      for _, automation in ipairs(src_track.automation) do
      
        -- Calculate scaled points.
        local points = {}
        local max_time = dst_length + 1
        for _, point in ipairs(automation.points) do
          local new_time = 1 + ((point.time - 1) * ratio)
          if (new_time < max_time) then
            table.insert(points, { 
              time = new_time,
              value = point.value 
            })
          end
        end
        local num_points = #points
        
        -- Clear old automation.
        automation:clear()
        
        -- Ensure the first and last points always get added.
        automation:add_point_at(points[1].time, points[1].value)
        if (num_points > 1) then
          automation:add_point_at(
            points[num_points].time, 
            points[num_points].value
          )
        end
        
        -- Add the in-between points.
        if (num_points > 2) then
          for i = 2, num_points - 1 do
            local point = points[i]
            if (not automation:has_point_at(point.time)) then
              automation:add_point_at(point.time, point.value)
            end
          end
        end
        
      end
    
    -- Else, we have no automation to process. 
    -- We can save time and just use copy_from()
    else
      
      src_track:copy_from(dst_track)
    
    end
       
    -- Clear temporary track for re-use.
    dst_track:clear()
     
    -- Show the delay column if any note delays have been used.
    if (rs:track(src_track_index).type == TRACK_TYPE_SEQUENCER) then
      if (delay_column_used) then
        rs:track(src_track_index).delay_column_visible = true
      end
    end
               
  end -- End: Iterate through each track.
 
  -- Remove temporary track.
  rs:delete_track_at(1)
  
end

-- GUI.
local function show_dialog()

  if dialog and dialog.visible then
    dialog:show()
    return
  end

  vb = renoise.ViewBuilder()
  
  local DIALOG_MARGIN = renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN
  local DIALOG_SPACING = renoise.ViewBuilder.DEFAULT_DIALOG_SPACING
  local CONTROL_MARGIN = renoise.ViewBuilder.DEFAULT_CONTROL_MARGIN

  local valuebox_length = vb:valuebox {
    min = 1,
    max = MAX_PATTERN_LINES,
    value = prefs.pattern_length.value,
    tostring = function(value) 
      return ("%d"):format(value)
    end,
    tonumber = function(str) 
      return tonumber(str, 10)
    end,
    notifier = function(value) 
      prefs.pattern_length.value = value 
    end
  }
  
  local BUTTON_WIDTH = 38
  local BUTTON_HEIGHT = 24
  local BUTTON_COLOUR = {31, 31, 31}
  local BUTTON_COLOUR_SHRINK = {127, 31, 31}
  local BUTTON_COLOUR_EXPAND = {31, 127, 31}
    
  local content = vb:column {
    
    uniform = true,
    style = 'body',
    margin = DIALOG_MARGIN,
    spacing = DIALOG_SPACING,  
        
    vb:column {

      style = 'body',
      margin = 10,
      spacing = 0,
      
      vb:row {
        spacing = 10,
        vb:text {
          font = 'bold',
          text = 'New Pattern Length',
        },
        valuebox_length,       
      },
      
    },
    
    vb:column {

      style = 'body',
      margin = 10,
      spacing = 0,
           
      vb:row {
      
        vb:column {
                 
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_SHRINK,
            text = '1 / 2',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(1/2)
            end
          },
                           
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_SHRINK,
            text = '1 / 3',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(1/3)
            end
          },
          
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_SHRINK,
            text = '1 / 4',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(1/4)
            end
          },
          
        },
        
        vb:column {
        
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_SHRINK,
            text = '2 / 3',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(2/3)
            end
          },
          
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_SHRINK,
            text = '3 / 4',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(3/4)
            end
          },
          
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_SHRINK,
            text = '4 / 5',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(4/5)
            end
          },
          
        },

        vb:column {
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT * 3,
            color = BUTTON_COLOUR,
            text = '1 / 1',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(1)
            end
          },
        },
        
        vb:column {
        
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_EXPAND,
            text = '3 / 2',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(3/2)
            end
          },
          
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_EXPAND,
            text = '4 / 3',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(4/3)
            end
          },
          
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_EXPAND,
            text = '5 / 4',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(5/4)
            end
          },
          
        },
        
        vb:column {
                 
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_EXPAND,
            text = '2 / 1',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(2/1)
            end
          },
                           
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_EXPAND,
            text = '3 / 1',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(3/1)
            end
          },
          
          vb:button {
            width = BUTTON_WIDTH,
            height = BUTTON_HEIGHT,
            color = BUTTON_COLOUR_EXPAND,
            text = '4 / 1',
            notifier = function()
              valuebox_length.value = adjust_pattern_length(4/1)
            end
          },
          
        },
        
      },
       
    },
  
    vb:button {
      height = 32,
      text = 'Resize',
      notifier = function()
        resize_pattern(renoise.song().selected_pattern, valuebox_length.value)
      end
    },

  } 
  
  dialog = renoise.app():show_custom_dialog(tool_name .. ' v' .. tool_version, content)  
end

renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:Illformed:Pattern Resizer",
  invoke = function()
    show_dialog()
  end
}

renoise.tool():add_menu_entry {
  name = "Pattern Editor:Pattern:Resize",
  invoke = function()
    show_dialog()
  end
}

renoise.tool():add_keybinding {
  name = "Pattern Editor:Pattern:Resize",
  invoke = function()
    show_dialog()
  end
}

-- Debug function.
_AUTO_RELOAD_DEBUG = function()
  show_dialog()
end
